#include <stdio.h>

#include "sage++user.h"

#include "anls_att.h"
#include "detect_comm.h"

extern char * mixstring;


extern int pcrc_detect_communication(detect_DIM* dad_x, detect_DIM* dad_y,
                              int* l, int* u, int* stride, int *a0, int *b0,
                              int* a1, int* b1, int *amount);

extern int getreflevel(SgSymbol * index, SgExpression * expr);
extern SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr);

void get_forall_expression(SgStatement * s, SgExpression * expr, Position side)
{
  if(expr) {
    switch(expr->variant()) {
    case INT_VAL:
      break;
    
    case VAR_REF: {
      forall_reference * fref= new forall_reference(side);
      fref->setExpr(expr);
      s->addAttribute(Forall_Ref,(void *)fref, sizeof(forall_reference));
      break;
    };

    case ARRAY_REF: {
      forall_reference * fref= new forall_reference(side);
      fref->setExpr(expr);
      s->addAttribute(Forall_Ref,(void *)fref, sizeof(forall_reference));
      break;
    };

    case MINUS_OP:
    case ADD_OP:    
    case NOTEQL_OP:
    case GT_OP: {
      get_forall_expression(s, expr->operand(1), side);
      get_forall_expression(s, expr->operand(2), side);
      break;
    }

    default:
      printf("Not touched expression in FORALL statement!\n");

    };
  };
};

void get_forall_reference(SgStatement * s)
{ 
  SgStatement * body = s->childList1(0);
    
  switch(body->variant()) {
  case ASSIGN_STAT: 
    get_forall_expression(body, body->expr(0), left);
    get_forall_expression(body, body->expr(1), right);
    break;

  case CONTROL_END: 
    break;
    
  default:
    printf("Illegal FORALL statement!\n");
    break;
  };

  Forall_header_def * fhdef=(Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  SgExpression * condition = fhdef->getcondition();

  if (condition!=NULL)
    get_forall_expression(body, condition, head);

};

int broadcasting_index_pattern(forall_reference * lhsref, forall_reference * rhsref)
{
  //if left has more index than the right;
  //if less, something wrong;
  // return (1);
  return (0);
}

int get_partitioned_dim(Processor_def * pdef, int dim, SgArrayRefExp * array)
{
   Alignment_def * adef = ((Alignment_def *) ((array->symbol()->getAttribute(0, Align_Def))->getAttributeData()) );
   SgSymbol * tsym = adef->gettosymbol();
   Template_def * tdef = (Template_def *) (tsym->getAttribute(0, Temp_Def)->getAttributeData());
   Distribute_def * ddef = (Distribute_def *) (tsym->getAttribute(0, Dist_Def)->getAttributeData());

   int dimcounter=0;

   for (int i=0; i<array->numberOfSubscripts(); i++) {
    if (ddef->distCode(i)!=NOP)
      if (dimcounter==dim)
	break;
      else
	dimcounter++;
   }	  
   return dimcounter;
}

void fill_DIM(forall_reference * ref, int dim, detect_DIM & dimstruc)
{
  SgSymbol * arraysym = ref->getexpression()->symbol();
  Alignment_def * adef = (Alignment_def *) ((arraysym->getAttribute(0, Align_Def))->getAttributeData());

  dimstruc.t_stride=adef->getStride(dim);
  dimstruc.t_offset=adef->getOffset(dim);

  SgSymbol * templatesym = adef->gettosymbol();
  Template_def * tdef = (Template_def *) ((templatesym->getAttribute(0, Temp_Def))->getAttributeData());

  dimstruc.t_lb=tdef->getStart(dim);
  dimstruc.t_ub=tdef->getEnd(dim);

  Distribute_def * ddef = (Distribute_def *) templatesym->getAttribute(0, Dist_Def)->getAttributeData();

  switch(ddef->distCode(dim)) {
  case NOP:
    dimstruc.distCode=0;
    break;
  case BLOCK:
    dimstruc.distCode=1;
    break;
  case CYCLIC:
    dimstruc.distCode=2;
    break;
  }

  // at present time, we consider distribute is automatically maped to P;

  if (dimstruc.distCode==NOP)
    dimstruc.numProcs=0;
  else {
    int nopnum=0;
    for (int i=0; i<dim; i++) {
      if (ddef->distCode(dim)==NOP) 
	nopnum++;
    };

    dim-=nopnum;

    SgSymbol * processorsym = ddef->gettosymbol();
    Processor_def * pdef = (Processor_def *) ((processorsym->getAttribute(0, Proc_Def))->getAttributeData());

    dimstruc.numProcs=(pdef->getEnd(dim))-(pdef->getStart(dim))+1;
  };
};


int forall_carry_dependent(forall_reference * leftref, 	forall_reference * rightref) {	 
  if (leftref->getexpression()->symbol()==rightref->getexpression()->symbol()) {

    // here we use a simple method, if the correspoding stride and offset is not equal;
    // a dependence is marked;

    for (int i=0; i< MAX_ALIGN_DIM; i++) {
      for (int j=0; j< MAX_INDEX_DIM; j++) {
	if  (leftref->getstride(i, j)!=NULL) {
	  if (rightref->getstride(i, j)!=NULL) {
	    if (leftref->getstride(i,j)->valueInteger()!=rightref->getstride(i,j)->valueInteger())
	      return (1);
	    else if (leftref->getoffset(i,j)->valueInteger()!=rightref->getoffset(i,j)->valueInteger())
	      return (1);
	  }
	  else
	    return (1);
	}
      }
    }
  }
  return (0);
}

void detect_communication(SgStatement * s, int counter)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();
  
  SgStatement * body = s->childList1(0);

  forall_reference * lhsref = (forall_reference *) ((body->getAttribute(0, Forall_Ref))->getAttributeData()); 

  forall_reference * rhsref = (forall_reference *) ((body->getAttribute(counter, Forall_Ref))->getAttributeData());

  SgArrayRefExp * lhsarray = (SgArrayRefExp *) lhsref->getexpression();
  SgSymbol * lhssymb = lhsarray->symbol();

  SgArrayRefExp * rhsarray = (SgArrayRefExp *) rhsref->getexpression();
  SgSymbol * rhssymb = rhsarray->symbol();

  if (lhssymb->numberOfAttributes(Align_Def)!=0) {
    Alignment_def * ladef = ((Alignment_def *) ((lhssymb->getAttribute(0, Align_Def))->getAttributeData()) );
    if (rhssymb->numberOfAttributes(Align_Def)!=0) {
      Alignment_def * radef = ((Alignment_def *) ((rhssymb->getAttribute(0, Align_Def))->getAttributeData()) );
      
      if (broadcasting_index_pattern(lhsref, rhsref)) {
	rhsref->setCommunication(BROADCAST);
	return;
      }
      else { 
	// try to decide if we can use shift, which is better than remap.

	SgSymbol * lt = ladef->gettosymbol();
	Template_def * ltdef = (Template_def *) (lt->getAttribute(0, Temp_Def)->getAttributeData());
	Distribute_def * lddef = (Distribute_def *) (lt->getAttribute(0, Dist_Def)->getAttributeData());

	SgSymbol * rt = radef->gettosymbol();
	Template_def * rtdef = (Template_def *) (rt->getAttribute(0, Temp_Def)->getAttributeData());
	Distribute_def * rddef = (Distribute_def *) (rt->getAttribute(0, Dist_Def)->getAttributeData());
      
	if (lddef->gettosymbol()!=rddef->gettosymbol()) 
	  printf("Data aligned to different processors!\n");
	else {
	  Processor_def * pdef = (Processor_def *) (lddef->gettosymbol()->getAttribute(0, Proc_Def)->getAttributeData());

	  for (int i=0; i<pdef->getDim(); i++) {
	    
	    //  if the partitioned array dim are both scalar or both in assign section,
	    //    do pcrc_detect_communication;
	    //  else 
	    //    do remap;

	    int lpd=get_partitioned_dim(pdef,i, lhsarray);
	    int rpd=get_partitioned_dim(pdef,i, rhsarray);

	    if (lhsref->scalarDim(lpd)&&rhsref->scalarDim(rpd))
	      printf("not finished!\n");

	    else if (lhsref->tripletDim(lpd)&&rhsref->tripletDim(rpd)) {
	      if (lhsref->indexNumberOnDim(lpd)!=rhsref->indexNumberOnDim(rpd)) {
		rhsref->setCommunication(REMAP);
		return;
	      }
	      else {
		if ((!lhsref->hasScalarDim())&&(!rhsref->hasScalarDim())) {
		  int amount;

		  int lb, ub, stride;

		  lb=fhdef->getindextriple(lhsref->indexNumberOnDim(lpd), 0)->valueInteger();
		  ub=fhdef->getindextriple(lhsref->indexNumberOnDim(lpd), 1)->valueInteger();

		  if (fhdef->getindextriple(lhsref->indexNumberOnDim(lpd), 2)!=NULL)
		    stride=fhdef->getindextriple(lhsref->indexNumberOnDim(lpd), 2)->valueInteger();
		  else
		    stride=1;

		  detect_DIM leftDIM, rightDIM;

		  int ls, lo, rs, ro;
	
		  fill_DIM(lhsref, lpd, leftDIM);
	    
		  if (lhsref->getstride(lpd, lhsref->indexNumberOnDim(lpd))!=NULL)
		    ls=lhsref->getstride(lpd, lhsref->indexNumberOnDim(lpd))->valueInteger();
		  else
		    ls=0;
		  if (lhsref->getoffset(lpd, lhsref->indexNumberOnDim(lpd))!=NULL)
		    lo=lhsref->getoffset(lpd, lhsref->indexNumberOnDim(lpd))->valueInteger();
		  else
		    lo=0;

		  fill_DIM(rhsref, rpd, rightDIM);
	    
		  if (rhsref->getstride(rpd, rhsref->indexNumberOnDim(rpd))!=NULL)
		    rs=rhsref->getstride(rpd, rhsref->indexNumberOnDim(rpd))->valueInteger();
		  else
		    rs=0;
		  if (rhsref->getoffset(rpd, rhsref->indexNumberOnDim(rpd))!=NULL)
		    ro=rhsref->getoffset(rpd, rhsref->indexNumberOnDim(rpd))->valueInteger();
		  else
		    ro=0;
		 
		  switch (pcrc_detect_communication(&leftDIM,&rightDIM,&lb,&ub,&stride,&ls,&lo,&rs,&ro,&amount)) {
		  case 0:
		  case 1:
		    rhsref->setCommunication(SHIFT);
		    rhsref->setAmount(rpd, amount);
		    break;
		  case 2:
		    rhsref->setCommunication(REMAP);
		    return;
		    break;
		  default:
		    printf("wrong out put\n");
		  
		  };
		}
		else {
		  // reduce dimmention;
		}

	      }
	    }
	    else {
	      rhsref->setCommunication(REMAP);
	      return;
	    }
	  }
	}
      }
    }
    else {
      if (rhssymb->numberOfAttributes(Inherit_Def)!=0) {
	rhsref->setCommunication(UNKNOW);
	return;
      }
      else {
	rhsref->setCommunication(NONE);
	return;
      }
    }
  }
  else {  // left is a non explicit distributed array;
    if (lhssymb->numberOfAttributes(Inherit_Def)!=0) {
      rhsref->setCommunication(UNKNOW);
      return;
    }
    else { // left is a non explicit distributed local array;
      if (rhssymb->numberOfAttributes(Align_Def)!=0) {

	// Only if all dim are duplicated, there is no communication;

	if (broadcasting_index_pattern(lhsref, rhsref))
	  rhsref->setCommunication(BROADCAST);
	else 
	  rhsref->setCommunication(REMAP);
	return;
      }
      else { 
	if (rhssymb->numberOfAttributes(Inherit_Def)!=0) {
	  rhsref->setCommunication(UNKNOW);
	  return;
	}
	else {
	  rhsref->setCommunication(NONE);
	  return;
	}
      }
    }
  }

  // check shift zero;

  if (rhsref->getCommunication()==SHIFT) {
    for (int i=0; i<MAX_ALIGN_DIM; i++)
      if (rhsref->getAmount(i)!=0)
	return;
    
    rhsref->setCommunication(NONE);
    return;
  }

  // check forall execution order;

  if (rhsref->getCommunication()==NONE) {
    if (forall_carry_dependent(lhsref, rhsref)) {
      rhsref->setCommunication(REMAP);
      return;
    }
  }
}

void detect_forall_reference(SgStatement * s)
{
  int header_evaluatable=1;

  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  for (int i=0; i<fhdef->getindexnumber(); i++) 
    header_evaluatable *= fhdef->evaluatable_index(i);

  SgStatement * body=s->childList1(0);
  
  switch(body->variant()) {
  case ASSIGN_STAT: 
    {
      int i;

      for (i=0; i<body->numberOfAttributes(Forall_Ref); i++) {
	forall_reference * fref = (forall_reference *) ((body->getAttribute(i, Forall_Ref))->getAttributeData());

	if (!fref->checkRegular()) {
	  fref->setCommunication(IRREG);
	  printf("Irregular forall reference!\n");
	}
      };
      
      forall_reference * lhsref = (forall_reference *) ((body->getAttribute(0, Forall_Ref))->getAttributeData()); 

      if (lhsref->getexpression()->variant()!=ARRAY_REF)
	printf("Multi assignment!\n");

      for (i=1; i<body->numberOfAttributes(Forall_Ref); i++) {
	forall_reference * rhsref = (forall_reference *) ((body->getAttribute(i, Forall_Ref))->getAttributeData());
	if (rhsref->getexpression()->variant()==ARRAY_REF) {
	  if (lhsref->evaluatable()) {
	    if (rhsref->evaluatable()) {
	      if (header_evaluatable)
		detect_communication(s, i);
	      else
		rhsref->setCommunication(REMAP);
	    }
	    else
	      rhsref->setCommunication(REMAP);
	  }
	  else
	    rhsref->setCommunication(REMAP);
	}
      };

      break;
    };
  case CONTROL_END: 
    break;    
  default:
    printf("Not touched yet!!!\n");
    break;
  };
};

void show_forall_reference(SgStatement * s)
{
  s=s->childList1(0);

  switch(s->variant()) {
  case ASSIGN_STAT: 
    {
      int i;
      for (i=0; i<s->numberOfAttributes(Forall_Ref); i++) {
	((forall_reference *) ((s->getAttribute(i, Forall_Ref))->getAttributeData())) ->show();
      };
    };
    break;

  case CONTROL_END: 
    break;
    
  default:
    printf("Not touched yet!!\n");
    break;
  };
};

void index_forall_reference(int indexdim, SgSymbol * index, SgStatement * s)
{
  switch(s->variant()) {
  case ASSIGN_STAT:     {
    for (int i=0; i<s->numberOfAttributes(Forall_Ref); i++) {
      forall_reference * fref= (forall_reference *) ((s->getAttribute(i, Forall_Ref))->getAttributeData());
      
      int indexlevel;
      
      SgExpression * frexp=fref->getexpression();
      
      switch(frexp->variant()) {
      case VAR_REF:
	indexlevel=getreflevel(index, frexp);
	fref->setIndex(0, indexdim, indexlevel);	  
	break;
      case ARRAY_REF:	  {
	SgArrayRefExp * arrayexp=(SgArrayRefExp *) frexp;
	
	SgSymbol * arraysym=arrayexp->symbol();
	
	for (int dimI=0; dimI<arrayexp->numberOfSubscripts();dimI++) {
	  indexlevel=getreflevel(index, arrayexp->subscript(dimI));
	  fref->setIndex(dimI, indexdim, indexlevel);
	  
	  // if possible, try to get stride and offset;
	  
	  if (indexlevel==1) {
	    SgExpression * stride = getlinerexp(1, index, arrayexp->subscript(dimI));
	    SgExpression * offset = getlinerexp(0, index, arrayexp->subscript(dimI));
	    
	    fref->setstride(dimI, indexdim, stride);
	    fref->setoffset(dimI, indexdim, offset);
	  }
	}
	
	break;
      };
      
      default:
	printf("Expression not touched!\n");
	break;
      };
    };
    break;
  };

  case CONTROL_END: 
    break;
    
  default:
    printf("Statement not touched yet!!!\n");
    break;
  };
};

//---------------------

void add_triple(SgExpression * e, Forall_header_def * fhdef)
{
  if (e->operand(2)!=NULL)
    fhdef->setindextriple(2, e->operand(2));
  fhdef->setindextriple(0, e->operand(1)->operand(1));
  fhdef->setindextriple(1, e->operand(1)->operand(2));
};

void add_index(SgExpression * e, Forall_header_def * fhdef)
{
  //printf("The id is %d\n", e->id());
  fhdef->setindexsymb(e->operand(1)->operand(1)->symbol());
  add_triple(e->operand(1)->operand(2), fhdef);

  if ((e->variant()==CONTROL_LIST) && (e->operand(2)!=NULL)) 
    add_index(e->operand(2), fhdef);
};

void add_forall_index(SgStatement * s, SgSymbol * symb)
{
  Forall_index_def * fidef = new Forall_index_def;
  fidef->setsymbol(symb);

  // add this attributes to both program header and symbol

  s->controlParent()->addAttribute(FINDEX_Def,(void *) fidef, sizeof(Forall_index_def));
  symb->addAttribute(FINDEX_Def,(void *) fidef, sizeof(Forall_index_def));

};

void analyse_forall_header(SgStatement * s)
{
  Forall_header_def * fhdef = new Forall_header_def;

  if (s->expr(1)!=NULL)
    fhdef->setcondition(s->expr(1));

  add_index(s->expr(0)->operand(1), fhdef);

  // add this attributes to forall statement
  s->addAttribute(FHDER_Def,(void *) fhdef, sizeof(Forall_header_def));

  // add this attributes to program header;
  s->controlParent()->addAttribute(FHDER_Def,(void *) fhdef, sizeof(Forall_header_def));

  for (int i=0; i<fhdef->getindexnumber(); i++) {
    SgSymbol * index=fhdef->getindexsymb(i);
      
    if(index->numberOfAttributes(FINDEX_Def)==0)
      add_forall_index(s, index);
  }
}

